#!/usr/bin/env python
# starfeeder/balance.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

NOTES:
Reference is [4].
Summary of commands on p14.

The balance doesn't have any concept of real-world mass, it seems.
So it needs calibrating.
We should NOT use the tare function, since the tare zero point is lost through
a reset. Instead, let's store it in the database.
"""

import datetime
import math
import re
from typing import Optional

import arrow
import bitstring
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot
import serial
from whisker.lang import CompiledRegexMemory
from whisker.qt import exit_on_exception

from starfeeder.constants import (
    BALANCE_ASF_MINIMUM,
    BALANCE_ASF_MAXIMUM,
    DEFAULT_BALANCE_READ_FREQUENCY_HZ,
    GUI_MASS_FORMAT,
)
from starfeeder.models import (
    BalanceConfig,  # for type hints
    CalibrationReport,
    MassEvent,
    RfidEvent,
    RfidReaderConfig,  # for type hints
)
from starfeeder.serial_controller import (
    # CR,
    CRLF,
    # LF,
    SerialController,
    SerialOwner,
)

# Startup sequence
CMD_NO_OP = ""  # p12: a termination character on its own clears the buffer
CMD_WARM_RESTART = "RES"  # p46; no reply; takes up to 3 s
CMD_SET_BAUD_RATE = "BDR"  # p17
CMD_QUERY_BAUD_RATE = "BDR?"  # p17
CMD_ASCII_RESULT_OUTPUT = "COF3"  # p19
CMD_DATA_DELIMITER_COMMA_CR_LF = "TEX172"  # set comma+CRLF as delimiter; p22

CMD_QUERY_IDENTIFICATION = "IDN?"  # p48
CMD_QUERY_STATUS = "ESR?"  # p58

# Signal processing and measurement; see summary on p6
CMD_SET_FILTER = "ASF"  # p37
CMD_FILTER_TYPE = "FMD"  # p37
CMD_DEACTIVATE_OUTPUT_SCALING = "NOV0"  # p31
# ... I can't get any NOV command (apart from the query, NOV?) to produce
# anything other than '?'.
CMD_QUERY_OUTPUT_SCALING = "NOV?"  # p31
CMD_TARE = "TAR"  # p41
CMD_MEASUREMENT_RATE = "ICR"  # p40; "Mv/s" = measured values per second
RATE_MAP_HZ_TO_CODE = {  # p40
    100: 0,
    50: 1,
    25: 2,
    12: 3,
    6: 4,
    3: 5,
    2: 6,
    1: 7,
}
CMD_QUERY_MEASURE = "MSV?"
# For example: MSV?10; gives 10 values.
# If you use COF11 then the "status" field isn't a countdown, it's something
# fixed. We don't have a simple countdown, so we have to maintain our own.
# And we can't say "keep going for ever", it seems.
CMD_STOP_MEASURING = "STP"  # p36

# Note also factory defaults on p50.
# Note also LED control, p57
# Examples of communication sequences/startup, p63.

RESET_PAUSE_MS = 3000  # p46
BAUDRATE_PAUSE_MS = 200  # p64 ("approx. 150ms"); p17 ("<15 ms")

RESPONSE_UNKNOWN = '?'
RESPONSE_NONSPECIFIC_OK = '0'
BAUDRATE_REGEX = re.compile(r"^(\d+),(\d)$")  # e.g. 09600,1

COMMAND_SEPARATOR = b";"

# NONSENSE # MASS_REGEX = re.compile(r"^(.*)\s+(\w+)$")  # e.g. "99.99 g" [5]
# ... the balance doesn't actually work out any mass for you.


class BalanceController(SerialController):  # separate controller thread
    mass_received = pyqtSignal(MassEvent)
    calibrated = pyqtSignal(CalibrationReport)

    def __init__(self,
                 balance_config: BalanceConfig,
                 reader_config: RfidReaderConfig,
                 rfid_effective_time_s: float,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.balance_config = balance_config
        self.reader_config = reader_config
        self.rfid_effective_time_s = rfid_effective_time_s

        # The cycle time should be <1 s so the user can ping/tare with a small
        # latency (we don't take the risk of interrupting an ongoing
        # measurement cycle for that). Roughly.
        # So, for example, at 6 Hz we could have 5 per cycle.
        self.measurements_per_batch = math.ceil(
            0.5 * self.balance_config.measurement_rate_hz)  # type: int
        self.command_queue = []
        self.n_pending_measurements = 0
        self.rfid_event_rfid = None
        self.rfid_event_expires = arrow.now()
        self.max_value = 100000  # default (NOV command) is 100,000
        self.recent_measurements_kg = []
        self.pending_calibrate = False
        self.pending_tare = False
        self.locked = False

    @pyqtSlot()
    @exit_on_exception
    def on_start(self) -> None:
        self.reset()
        self.check_calibrated()

    def send(self, command: str, params: str = '', reply_expected: bool = True,
             delay_ms: int = 0) -> None:
        params = str(params)  # just in case we have a number
        if reply_expected:
            self.command_queue.append(command)
        msg = command + params
        super().send_str(msg, delay_ms)

    def check_calibrated(self) -> None:
        if self.value_to_mass(1) is None:
            self.warning("Balance uncalibrated; will not read yet")

    def reset(self) -> None:
        self.info("Balance resetting")
        self.recent_measurements_kg = []
        baud = self.balance_config.baudrate
        parity = self.balance_config.parity
        if parity == serial.PARITY_NONE:
            parity_code = 0
        elif parity == serial.PARITY_EVEN:
            parity_code = 1
        else:
            self.error("Invalid parity ({})! Choosing even parity. "
                       "COMMUNICATION MAY BREAK.")
            parity_code = 1

        self.send(CMD_NO_OP, reply_expected=False)  # cancel anything ongoing
        self.send(CMD_STOP_MEASURING, reply_expected=False)
        self.send(CMD_WARM_RESTART, reply_expected=False)
        # We want a short pause before sending the next command... but we can
        # line it up safely with this:.
        self.send(CMD_SET_BAUD_RATE, "{},{}".format(baud, parity_code),
                  delay_ms=RESET_PAUSE_MS)
        # Likewise a pause next (organized by the SerialWriter)...
        self.send(CMD_QUERY_BAUD_RATE, delay_ms=BAUDRATE_PAUSE_MS)
        self.send(CMD_QUERY_IDENTIFICATION)
        self.send(CMD_QUERY_STATUS)
        self.send(CMD_ASCII_RESULT_OUTPUT)
        self.send(CMD_DATA_DELIMITER_COMMA_CR_LF)
        # self.send(CMD_DEACTIVATE_OUTPUT_SCALING)  # Not working
        self.send(CMD_QUERY_OUTPUT_SCALING)
        asf = self.balance_config.amp_signal_filter_mode
        if asf < BALANCE_ASF_MINIMUM or asf > BALANCE_ASF_MAXIMUM:
            self.warning("Bad ASF mode ignored: {}".format(asf))
        else:
            self.send(CMD_SET_FILTER, asf)
        # noinspection PyTypeChecker
        self.send(CMD_FILTER_TYPE,
                  1 if self.balance_config.fast_response_filter else 0)
        # noinspection PyTypeChecker
        if self.balance_config.measurement_rate_hz not in RATE_MAP_HZ_TO_CODE:
            self.error(
                "Measurement rate {} is invalid; defaulting to {}".format(
                    self.balance_config.measurement_rate_hz,
                    DEFAULT_BALANCE_READ_FREQUENCY_HZ))
            frequency_code = RATE_MAP_HZ_TO_CODE.get(
                DEFAULT_BALANCE_READ_FREQUENCY_HZ)
        else:
            frequency_code = RATE_MAP_HZ_TO_CODE.get(
                self.balance_config.measurement_rate_hz)
        self.send(CMD_MEASUREMENT_RATE, str(frequency_code))
        self.start_measuring()

    def start_measuring(self) -> None:
        self.n_pending_measurements += self.measurements_per_batch
        # noinspection PyTypeChecker
        self.send(CMD_QUERY_MEASURE, self.measurements_per_batch)
        self.command_queue.extend(
            [CMD_QUERY_MEASURE] * (self.measurements_per_batch - 1))

    def read_until(self, when_to_read_until: arrow.Arrow) -> None:
        self.rfid_event_expires = when_to_read_until
        now = arrow.now()
        if self.balance_config.read_continuously:
            return  # already measuring
        if self.n_pending_measurements == 0 and now < self.rfid_event_expires:
            self.start_measuring()

    @pyqtSlot()
    @exit_on_exception
    def tare(self) -> None:
        # Don't use hardware tare:
            # # Commands (and output) are neatly queued up behind MSV commands.
            # # So this will happen at the end of the current MSV cycle.
            # self.send(CMD_TARE)
        # Use software tare.
        self.pending_tare = True
        if self.n_pending_measurements == 0:
            self.start_measuring()

    @pyqtSlot()
    @exit_on_exception
    def calibrate(self) -> None:
        self.pending_calibrate = True
        if self.n_pending_measurements == 0:
            self.start_measuring()

    @pyqtSlot()
    @exit_on_exception
    def ping(self) -> None:
        self.check_calibrated()
        # Commands (and output) are neatly queued up behind MSV commands.
        # So this will happen at the end of the current MSV cycle.
        self.status("Asking balance for identification")
        self.send(CMD_QUERY_IDENTIFICATION)
        self.status("Asking balance for status")
        self.send(CMD_QUERY_STATUS)

    def stop_measuring(self):
        self.send(CMD_STOP_MEASURING, reply_expected=False)
        self.command_queue = [x for x in self.command_queue
                              if x != CMD_QUERY_MEASURE]
        self.n_pending_measurements = 0

    @pyqtSlot()
    @exit_on_exception
    def on_stop(self) -> None:
        self.stop_measuring()
        self.finished.emit()
        # Inelegant! Risk the writer thread will be terminated before it
        # sends this command. Still, ho-hum.

    def report_status(self) -> None:
        self.status("Currently scanning" if self.n_pending_measurements > 0
                    else "Not currently scanning")

    def value_to_mass(self, value: int) -> Optional[float]:
        r = self.balance_config.refload_value
        z = self.balance_config.zero_value
        m = self.balance_config.refload_mass_kg
        if None in [r, z, m]:
            return None
        return m * (value - z) / (r - z)

    def is_stable(self) -> bool:
        # Do we have a stable mass?
        if len(self.recent_measurements_kg) < self.balance_config.stability_n:
            # ... not enough measurements to judge
            return False
        min_kg = min(self.recent_measurements_kg)
        max_kg = max(self.recent_measurements_kg)
        range_kg = max_kg - min_kg
        if range_kg > self.balance_config.tolerance_kg:
            return False
        # Stable.
        return True

    def process_value(self, value: int, timestamp: arrow.Arrow) -> None:
        mass_kg = self.value_to_mass(value)
        if mass_kg is None:
            self.debug("Balance uncalibrated; ignoring value")
            return
        self.debug("BALANCE VALUE: {} => {} kg".format(
            value, GUI_MASS_FORMAT % mass_kg))
        while (len(self.recent_measurements_kg) >=
                self.balance_config.stability_n):
            self.recent_measurements_kg.pop(0)
        self.recent_measurements_kg.append(mass_kg)
        rfid_valid = timestamp < self.rfid_event_expires
        rfid = self.rfid_event_rfid if rfid_valid else None
        identified = rfid is not None
        stable = self.is_stable()
        stable_high = stable and mass_kg >= self.balance_config.min_mass_kg
        stable_low = stable and mass_kg <= self.balance_config.unlock_mass_kg
        lock_now = stable_high and identified and not self.locked
        unlock_now = stable_low and self.locked
        mass_event = MassEvent(
            balance_id=self.balance_config.id,
            balance_name=self.balance_config.name,
            reader_id=self.reader_config.id,
            reader_name=self.reader_config.name,
            rfid=rfid,
            mass_kg=mass_kg,
            timestamp=timestamp,
            stable=stable,
            locked=lock_now,
        )
        if lock_now:
            self.locked = True
            self.debug("LOCKING at {}".format(mass_kg))
        elif unlock_now:
            self.locked = False
            self.debug("UNLOCKING at {}".format(mass_kg))
        self.mass_received.emit(mass_event)

    def tare_with(self, value: int) -> None:
        self.debug("tare_with: {}".format(value))
        self.pending_tare = False
        if value == self.balance_config.zero_value:
            return  # No change
        if self.balance_config.zero_value is None:
            self.balance_config.zero_value = value
        else:
            delta = value - self.balance_config.zero_value
            self.balance_config.zero_value += delta
            if self.balance_config.refload_value is not None:
                self.balance_config.refload_value += delta
        self.finish_calibration()

    def calibrate_with(self, value: int) -> None:
        self.debug("calibrate_with: {}".format(value))
        self.pending_calibrate = False
        if value == self.balance_config.refload_value:
            return  # no change
        self.balance_config.refload_value = value
        self.finish_calibration()

    def finish_calibration(self) -> None:
        if self.balance_config.refload_value == self.balance_config.zero_value:
            # would cause division by zero
            self.balance_config.refload_value = None
        report = CalibrationReport(
            balance_id=self.balance_config.id,
            balance_name=self.balance_config.name,
            zero_value=self.balance_config.zero_value,
            refload_value=self.balance_config.refload_value,
            refload_mass_kg=self.balance_config.refload_mass_kg,
        )
        self.calibrated.emit(report)
        self.check_calibrated()

    @pyqtSlot(RfidEvent)
    @exit_on_exception
    def on_rfid(self, rfid_event: RfidEvent) -> None:
        if not isinstance(rfid_event, RfidEvent):
            self.critical("Bad rfid_event: {}".format(rfid_event))
            return
        rfid_event_expires = rfid_event.timestamp + datetime.timedelta(
            seconds=self.rfid_effective_time_s)  # type: arrow.Arrow
        # ... an Arrow plus a timedelta gives an Arrow
        self.rfid_event_rfid = rfid_event.rfid
        self.read_until(rfid_event_expires)

    @pyqtSlot(str, arrow.Arrow)
    def on_receive(self, data: str, timestamp: arrow.Arrow) -> None:
        if not isinstance(data, str):
            self.critical("bad data: {}".format(repr(data)))
            return
        if not isinstance(timestamp, arrow.Arrow):
            self.critical("bad timestamp: {}".format(repr(timestamp)))
            return
        gre = CompiledRegexMemory()
        if self.command_queue:
            cmd = self.command_queue.pop(0)
        else:
            cmd = None
        self.debug("Balance receiving at {}: {} (most recent command was: "
                   "{})".format(timestamp, repr(data), cmd))
        # self.debug("len(self.command_queue) %d" % len(self.command_queue))

        if cmd == CMD_QUERY_MEASURE:
            try:
                value = int(data)
                if self.pending_tare:
                    self.tare_with(value)
                elif self.pending_calibrate:
                    self.calibrate_with(value)
                else:
                    self.process_value(value, timestamp)
            except ValueError:
                self.error("Balance sent a bad value")
            self.n_pending_measurements -= 1
            self.debug("n_pending_measurements: {}".format(
                self.n_pending_measurements))
            if (self.n_pending_measurements == 0 and
                    (self.balance_config.read_continuously or
                        self.locked or
                        timestamp < self.rfid_event_expires)):
                self.debug("Finished measuring; restarting")
                self.start_measuring()
        elif (cmd in [CMD_QUERY_BAUD_RATE, CMD_SET_BAUD_RATE] and
                gre.match(BAUDRATE_REGEX, data)):
            baudrate = int(gre.group(1))
            parity_code = int(gre.group(2))
            if parity_code == 1:
                parity = 'E'
            elif parity_code == 0:
                parity = 'N'
            else:
                parity = '?'
            self.status("Balance is using {} bps, parity {}".format(baudrate,
                                                                    parity))
        elif (data == RESPONSE_NONSPECIFIC_OK and
                cmd in [
                    CMD_ASCII_RESULT_OUTPUT,
                    CMD_DATA_DELIMITER_COMMA_CR_LF,
                    CMD_FILTER_TYPE,
                    CMD_MEASUREMENT_RATE,
                    CMD_SET_BAUD_RATE,
                    CMD_SET_FILTER,
                    CMD_TARE,
                ]):
            self.status("Balance acknowledges command {}".format(cmd))
        elif cmd == CMD_QUERY_STATUS:
            self.status("Balance status: {}".format(data))
            # noinspection PyBroadException
            try:
                bits = bitstring.BitStream(uint=int(data), length=6)
                command_error, execution_error, hardware_error = (
                    bits.readlist('bool, bool, bool, pad:3'))
                self.status(
                    "command_error={}, execution_error={}, "
                    "hardware_error={}".format(command_error, execution_error,
                                               hardware_error))
            except:
                self.status("Can't interpret status")
        elif cmd == CMD_QUERY_IDENTIFICATION:
            self.status("Balance identification: {}".format(data))
        elif cmd == CMD_QUERY_OUTPUT_SCALING:
            try:
                self.max_value = int(data)
            except ValueError:
                self.error("Bad value received")
        elif data == RESPONSE_UNKNOWN:
            self.status("Balance says 'eh?'")
        else:
            self.error(
                "Unknown message from balance: {} (for command: {})".format(
                    repr(data), repr(cmd)))


class BalanceOwner(SerialOwner):  # GUI thread
    # Outwards, to world:
    mass_received = pyqtSignal(MassEvent)
    calibrated = pyqtSignal(CalibrationReport)
    # Inwards, to posessions:
    ping_requested = pyqtSignal()
    tare_requested = pyqtSignal()
    calibration_requested = pyqtSignal()
    on_rfid = pyqtSignal(RfidEvent)

    def __init__(self,
                 balance_config: BalanceConfig,
                 rfid_effective_time_s: float,
                 callback_id: int,
                 parent: QObject = None) -> None:
        # Do not keep a copy of balance_config; it will expire.
        super().__init__(
            serial_args=balance_config.get_serial_args(),
            parent=parent,
            callback_id=callback_id,
            name=balance_config.name,
            rx_eol=CRLF,
            tx_eol=COMMAND_SEPARATOR,
            controller_class=BalanceController,
            controller_kwargs=dict(
                balance_config=balance_config.get_attrdict(),
                reader_config=balance_config.reader.get_attrdict(),
                rfid_effective_time_s=rfid_effective_time_s,
            ))
        # Balance uses CR+LF terminator when sending to computer [4].
        # Computer can use LF or semicolon terminator when sending to balance
        # [4].
        # Sometimes a semicolon is accepted but LF isn't (p23), so we use a
        # semicolon.
        self.balance_id = balance_config.id  # used by main GUI
        self.name = balance_config.name  # used by main GUI
        self.refload_mass_kg = balance_config.refload_mass_kg  # used by GUI (tare dialog)  # noqa
        self.ping_requested.connect(self.controller.ping)
        self.tare_requested.connect(self.controller.tare)
        self.calibration_requested.connect(self.controller.calibrate)

        self.on_rfid.connect(self.controller.on_rfid)  # different thread
        self.controller.mass_received.connect(self.mass_received)  # different thread  # noqa
        self.controller.calibrated.connect(self.calibrated)  # different thread

    def ping(self) -> None:
        self.ping_requested.emit()

    def tare(self) -> None:
        self.tare_requested.emit()

    def calibrate(self) -> None:
        self.calibration_requested.emit()



#!/usr/bin/env python
# starfeeder/getch.py

import atexit
import select
import sys

try:
    import msvcrt  # Windows only
    termios = None
    tty = None
except ImportError:
    msvcrt = None
    import termios  # Unix only
    import tty  # requires termios, so Unix only


# =============================================================================
# Read single character, waiting for it
# =============================================================================
# http://stackoverflow.com/questions/510357/python-read-a-single-character-from-the-user  # noqa
# http://home.wlu.edu/~levys/software/kbhit.py
# ... modified a little

def _getch_windows():
    """Gets a single character from standard input.  Does not echo to the
    screen."""
    return msvcrt.getch().decode('utf-8')


def _getch_unix():
    """Gets a single character from standard input.  Does not echo to the
    screen. Note that the terminal will have been pre-configured, below."""
    return sys.stdin.read(1)


# =============================================================================
# Is a keystroke available?
# =============================================================================
# http://code.activestate.com/recipes/572182-how-to-implement-kbhit-on-linux/
# http://stackoverflow.com/questions/2408560/python-nonblocking-console-input

def _kbhit_windows():
    return msvcrt.kbhit()


def _kbhit_unix():
    dr, dw, de = select.select([sys.stdin], [], [], 0)
    return dr != []


# =============================================================================
# Configure terminal (UNIX)
# =============================================================================

def set_normal_term():
    # switch to normal terminal
    termios.tcsetattr(_fd, termios.TCSAFLUSH, _old_term)


def set_curses_term():
    # switch to unbuffered terminal
    termios.tcsetattr(_fd, termios.TCSAFLUSH, _new_term)


# =============================================================================
# Set up for specific OS
# =============================================================================

if msvcrt:
    # -------------------------------------------------------------------------
    # Windows
    # -------------------------------------------------------------------------
    getch = _getch_windows
    kbhit = _kbhit_windows
else:
    # -------------------------------------------------------------------------
    # Unix
    # -------------------------------------------------------------------------
    getch = _getch_unix
    kbhit = _kbhit_unix

    # save the terminal settings
    _fd = sys.stdin.fileno()
    _new_term = termios.tcgetattr(_fd)
    _old_term = termios.tcgetattr(_fd)
    # new terminal setting unbuffered
    _new_term[3] = (_new_term[3] & ~termios.ICANON & ~termios.ECHO)
    _new_term[6][termios.VMIN] = 1
    _new_term[6][termios.VTIME] = 0

    atexit.register(set_normal_term)
    set_curses_term()



#!/usr/bin/env python
# starfeeder/main.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


REFERENCES cited in code:
[1] E-mail to Rudolf Cardinal from Søren Ellegaard, 9 Dec 2014.
[2] E-mail to Rudolf Cardinal from Søren Ellegaard, 10 Dec 2014.
[3] "RFID Reader.docx" in [1]; main reference for the RFID tag reader.
[4] "ba_ad105_e_2.pdf" in [1]; main reference for the balance.
[5] "RFID and LOAD CELL DEVICES - SE_20141209.pptx" in [1].
[6] E-mail to Rudolf Cardinal from Matthew Weinie, 8 Dec 2015.

"""

import argparse
import logging
import sys
import traceback

from PyQt5.Qt import PYQT_VERSION_STR
from PyQt5.QtCore import QT_VERSION_STR
from PyQt5.QtWidgets import QApplication
from sqlalchemy import create_engine
# from whisker.debug_qt import enable_signal_debugging_simply
from whisker.logging import (
    configure_logger_for_colour,
    copy_root_log_to_file,
)
from whisker.sqlalchemy import (
    get_current_and_head_revision,
    upgrade_database,
)
from whisker.qt import (
    GarbageCollector,
    LogWindow,
    run_gui,
)
import whisker.version

from starfeeder.constants import (
    ALEMBIC_BASE_DIR,
    ALEMBIC_CONFIG_FILENAME,
    DATABASE_ENV_VAR_NOT_SPECIFIED,
    DB_URL_ENV_VAR,
    LOG_FORMAT,
    LOG_DATEFMT,
    WINDOW_TITLE,
    WRONG_DATABASE_VERSION_STUB,
)
from starfeeder.gui import (
    BaseWindow,
    NoDatabaseSpecifiedWindow,
    WrongDatabaseVersionWindow,
)
from starfeeder.settings import (
    get_database_url,
    set_database_url,
    set_database_echo,
)
from starfeeder.version import VERSION

log = logging.getLogger(__name__)


# =============================================================================
# Main
# =============================================================================

def main() -> int:
    # -------------------------------------------------------------------------
    # Arguments
    # -------------------------------------------------------------------------
    parser = argparse.ArgumentParser(
        description="Starfeeder v{}. Whisker bird monitor, reading from RFID "
        "tag readers and weighing balances.".format(VERSION))
    # ... allow_abbrev=False requires Python 3.5
    parser.add_argument("--logfile", default=None,
                        help="Filename to append log to")
    parser.add_argument('--verbose', '-v', action='count', default=0,
                        help="Be verbose (use twice for extra verbosity)")
    parser.add_argument('--guilog', action="store_true",
                        help="Show Python log in a GUI window")
    parser.add_argument('--upgrade-database', action="store_true",
                        help="Upgrade database (determined from SQLAlchemy"
                        " URL, read from {} environment variable) to current"
                        " version".format(DB_URL_ENV_VAR))
    parser.add_argument('--gui', '-g', action="store_true",
                        help="GUI mode only")
    parser.add_argument(
        "--dburl", default=None,
        help="Database URL (if not specified, task will look in {} "
        "environment variable).".format(DB_URL_ENV_VAR))
    parser.add_argument('--dbecho', action="store_true",
                        help="Echo SQL to log.")
    parser.add_argument('--debug-qt-signals', action="store_true",
                        help="Debug QT signals.")

    # We could allow extra Qt arguments:
    # args, unparsed_args = parser.parse_known_args()
    # Or not:
    args = parser.parse_args()
    unparsed_args = []

    qt_args = sys.argv[:1] + unparsed_args

    # -------------------------------------------------------------------------
    # Modify settings if we're in a PyInstaller bundle
    # -------------------------------------------------------------------------
    in_bundle = getattr(sys, 'frozen', False)
    if in_bundle:
        args.gui = True
    if not args.gui:
        args.guilog = False

    # -------------------------------------------------------------------------
    # Create QApplication before we create any windows (or Qt will crash)
    # -------------------------------------------------------------------------
    qt_app = QApplication(qt_args)

    # Attempt to fix rare bugs by constraining the Python garbage collector
    # to the GUI thread only

    # noinspection PyUnusedLocal
    my_garbage_collector = GarbageCollector(qt_app, interval_ms=100)
    # ... with interval of 10000, get
    #     pymysql.err.OperationalError (1040, 'Too many connections')
    # ... 500 seems OK
    # ... not sure why there was a problem, though, as all sessions are created
    #     using "with session_thread_scope", and when that finishes, it calls
    #     session.close(), and that's meant to release all resources:
    #     http://docs.sqlalchemy.org/en/latest/orm/session_basics.html#closing
    # ... since I'm not sure, let's use 100 to be conservative

    # -------------------------------------------------------------------------
    # Logging
    # -------------------------------------------------------------------------
    loglevel = logging.DEBUG if args.verbose >= 1 else logging.INFO
    logging.basicConfig(format=LOG_FORMAT, datefmt=LOG_DATEFMT,
                        level=loglevel)
    rootlogger = logging.getLogger()
    rootlogger.setLevel(loglevel)
    configure_logger_for_colour(rootlogger)  # configure root logger
    logging.getLogger('whisker').setLevel(logging.DEBUG if args.verbose >= 2
                                          else logging.INFO)
    if args.logfile:
        copy_root_log_to_file(args.logfile)
    if args.guilog:
        log_window = LogWindow(level=loglevel,
                               window_title=WINDOW_TITLE + " Python log",
                               logger=rootlogger)
        log_window.show()

    # If any exceptions happen up to this point, we're a bit stuffed.
    # But from now on, we can trap anything and see it in the GUI log, if
    # enabled, even if we have no console.

    # noinspection PyBroadException
    try:

        # ---------------------------------------------------------------------
        # Info
        # ---------------------------------------------------------------------
        log.info("Starfeeder v{}: RFID/balance controller for Whisker, "
                 "by Rudolf Cardinal (rudolf@pobox.com)".format(VERSION))
        log.debug("args: {}".format(args))
        log.debug("qt_args: {}".format(qt_args))
        log.debug("PyQt5 version: {}".format(PYQT_VERSION_STR))
        log.debug("Qt version: {}".format(QT_VERSION_STR))
        log.debug("Whisker client version: {}".format(whisker.version.VERSION))
        if in_bundle:
            log.debug("Running inside a PyInstaller bundle")
        if args.gui:
            log.debug("Running in GUI-only mode")
        # if args.debug_qt_signals:
        #     enable_signal_debugging_simply()

        # ---------------------------------------------------------------------
        # Database
        # ---------------------------------------------------------------------
        # Get URL, or complain
        if args.dburl:
            set_database_url(args.dburl)
        if args.dbecho:
            set_database_echo(args.dbecho)
        try:
            database_url = get_database_url()
        except ValueError:
            if args.gui:
                win = NoDatabaseSpecifiedWindow()
                if args.guilog:
                    # noinspection PyUnboundLocalVariable
                    win.exit_kill_log.connect(log_window.exit)
                return run_gui(qt_app, win)
            raise ValueError(DATABASE_ENV_VAR_NOT_SPECIFIED)
        engine = create_engine(database_url)
        log.debug("Using database URL: {}".format(engine))  # obscures password

        # Has the user requested a command-line database upgrade?
        if args.upgrade_database:
            sys.exit(upgrade_database(ALEMBIC_CONFIG_FILENAME,
                                      ALEMBIC_BASE_DIR))
        # Is the database at the correct version?
        (current_revision, head_revision) = get_current_and_head_revision(
            database_url,
            ALEMBIC_CONFIG_FILENAME,
            ALEMBIC_BASE_DIR
        )
        if current_revision != head_revision:
            if args.gui:
                win = WrongDatabaseVersionWindow(current_revision,
                                                 head_revision)
                if args.guilog:
                    # noinspection PyUnboundLocalVariable
                    win.exit_kill_log.connect(log_window.exit)
                return run_gui(qt_app, win)
            raise ValueError(WRONG_DATABASE_VERSION_STUB.format(
                head_revision=head_revision,
                current_revision=current_revision))

        # ---------------------------------------------------------------------
        # Run app
        # ---------------------------------------------------------------------
        win = BaseWindow()
        if args.guilog:
            # noinspection PyUnboundLocalVariable
            win.exit_kill_log.connect(log_window.exit)
        return run_gui(qt_app, win)

    except:
        if args.guilog:
            log.critical(traceback.format_exc())
            log_window.set_may_close(True)
            return qt_app.exec_()
        else:
            raise


# =============================================================================
# Command-line entry point
# =============================================================================

if __name__ == '__main__':
    # noinspection PyBroadException
    try:
        sys.exit(main())
    except Exception as e:  # master top-level exception catcher
        log.critical("Exception caught at top level: {}".format(str(e)))
        log.critical(traceback.format_exc())
        # Don't use print_exc(); that might not get sent to the log.
        sys.exit(1)



#!/usr/bin/env python
# starfeeder/alembic/env.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
# import os
# import sys

from alembic import context
from sqlalchemy import engine_from_config, pool

from starfeeder.constants import LOG_FORMAT, LOG_DATEFMT
from starfeeder.models import Base
from starfeeder.settings import get_database_settings

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())
logging.basicConfig(format=LOG_FORMAT, datefmt=LOG_DATEFMT,
                    level=logging.DEBUG)

# current_dir = os.path.dirname(os.path.abspath(__file__))
# project_dir = os.path.abspath(os.path.join(current_dir,
#                                            os.pardir, os.pardir))
# log.info("Adding to PYTHONPATH: {}".format(project_dir))
# sys.path.append(project_dir)
# log.debug("sys.path: {}".format(sys.path))

config = context.config
target_metadata = Base.metadata
settings = get_database_settings()
config.set_main_option('sqlalchemy.url', settings['url'])


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    # RNC
    context.configure(
        url=url,
        target_metadata=target_metadata,
        render_as_batch=True,  # for SQLite mode; http://stackoverflow.com/questions/30378233  # noqa
        literal_binds=True,
    )
    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix='sqlalchemy.',
        poolclass=pool.NullPool)

    with connectable.connect() as connection:
        # RNC
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            render_as_batch=True,  # for SQLite mode; http://stackoverflow.com/questions/30378233  # noqa
        )
        with context.begin_transaction():
            context.run_migrations()

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



#!/usr/bin/env python
# setup.py

"""
Starfeeder setup file

"""
# http://python-packaging-user-guide.readthedocs.org/en/latest/distributing/
# http://jtushman.github.io/blog/2013/06/17/sharing-code-across-applications-with-python/  # noqa

from setuptools import setup, find_packages
from codecs import open
from os import path
# import pip
# from pip.req import parse_requirements

from starfeeder.version import VERSION

here = path.abspath(path.dirname(__file__))

# -----------------------------------------------------------------------------
# Get the long description from the README file
# -----------------------------------------------------------------------------
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# -----------------------------------------------------------------------------
# Get the requirements from the requirements file
# -----------------------------------------------------------------------------
# http://stackoverflow.com/questions/14399534
# https://github.com/juanpabloaj/pip-init/issues/11
# reqfile = path.join(here, 'requirements.txt')
# install_reqs = parse_requirements(reqfile, session=pip.download.PipSession())
# reqs = [str(ir.req) if ir.req else str(ir.link) for ir in install_reqs]
# ... RNC modification: for github ones, the .req is None but .link works
# ... no, we have to use fancy stuff for the github ones;
#     http://stackoverflow.com/questions/18026980

# -----------------------------------------------------------------------------
# setup args
# -----------------------------------------------------------------------------
setup(
    name='starfeeder',

    version=VERSION,

    description='Whisker Starfeeder (starling RFID/balance reader)',
    long_description=long_description,

    # The project's main homepage.
    url='http://www.whiskercontrol.com/',

    # Author details
    author='Rudolf Cardinal',
    author_email='rudolf@pobox.com',

    # Choose your license
    license='GNU General Public License v3 or later (GPLv3+)',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',  # noqa

        'Natural Language :: English',

        'Operating System :: OS Independent',

        'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.2',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',  # PyQt5 requires 3.5
        'Programming Language :: Python :: 3 :: Only',

        'Topic :: System :: Hardware',
        'Topic :: System :: Networking',
    ],

    keywords='whisker rfid weigh balance starling',

    packages=find_packages(),  # finds all the .py files in subdirectories
    package_data={
        'starfeeder': [
            'alembic.ini',
        ],
        'starfeeder.alembic': [
            'env.py',
        ],
        'starfeeder.alembic.versions': [
            '*.py',
        ],
    },

    install_requires=[
        # ---------------------------------------------------------------------
        # Standard PyPI packages
        # ---------------------------------------------------------------------
        'alembic>=0.8.3',  # migration tool for sqlalchemy
        'arrow',  # better datetime
        'attrdict',  # dictionaries with attribute-style access
        'bitstring>=3.1.4',  # manipulation of binary numbers
        # 'PySide==1.2.4',  # Python interface to Qt
        'PyQt5>=5.7',  # Python interface to Qt
        'pyserial>=3.2.1',  # serial port library
        'six>=1.10.0',  # Python 2/3 compatibility library
        'SQLAlchemy>=1.0.12',  # database ORM
        # 'typing==3.5.2.2',  # part of stdlib in Python 3.5, but not 3.4
        'whisker>=0.3.5',  # Whisker client library

        # ---------------------------------------------------------------------
        # Database connections (SQLite is built in)
        # ---------------------------------------------------------------------
        # MySQL:
        'PyMySQL>=0.7.2',

        # SQL Server / ODBC route:
        'pyodbc>=3.0.10',

        # PostgreSQL:
        'psycopg2>=2.6.1',

        # ---------------------------------------------------------------------
        # Specials: development versions
        # ---------------------------------------------------------------------
        # 'pyserial==3.0b1',

        # ---------------------------------------------------------------------
        # For development only
        # ---------------------------------------------------------------------
        # 'PyInstaller>=3.2',
        # 'colorama',
        # 'docutils==0.12',  # documentation generation tools
        # 'git+https://github.com/pyinstaller/pyinstaller@93d25e563e49f88f5977afd9b6e26cb34bfb5efa',  # PyInstaller=3.1.dev0+93d25e5  # noqa
    ],
    dependency_links=[
        # ---------------------------------------------------------------------
        # PySerial development version
        # ---------------------------------------------------------------------
        # We browse at https://github.com/pyserial/pyserial
        # We want the commit 3e02f7052747521a21723a618dccf303065da732
        # We want the tarball
        # The API is:
        #   GET /repos/:owner/:repo/:archive_format/:ref
        #   - https://developer.github.com/v3/repos/contents/#get-archive-link
        # or
        #   https://github.com/user/project/archive/commit.zip
        #   - http://stackoverflow.com/questions/17366784
        # or
        #   http://github.com/usr/repo/tarball/tag
        # That gets us:
        #   https://github.com/pyserial/pyserial/tarball/3e02f7052747521a21723a618dccf303065da732  # noqa
        # We label it with "#egg=pyserial-3.0b1" for setup.py's benefit
        #   http://stackoverflow.com/questions/3472430
        # Final answer:

        # 'http://github.com/pyserial/pyserial/tarball/3e02f7052747521a21723a618dccf303065da732#egg=pyserial-3.0b1',  # noqa

        # ---------------------------------------------------------------------
        # MySQL Connector/Python
        # ---------------------------------------------------------------------
        # 'https://cdn.mysql.com/Downloads/Connector-Python/mysql-connector-python-2.1.3.tar.gz',  # MySQL Connector/Python  # noqa
    ],
    # YOU MUST ALSO USE THE "--process-dependency-links" FLAG.

    entry_points={
        'console_scripts': [
            # Format is 'script=module:function".
            'starfeeder=starfeeder.main:main',
            'starfeeder_mimic_rfid_reader=starfeeder.test_mimic_rfid_reader:main',  # noqa
            'starfeeder_mimic_balance=starfeeder.test_mimic_balance:main',
        ],
    },
)



#!/usr/bin/env python
# starfeeder/serial_controller.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from collections import deque, OrderedDict
import logging
import platform
import traceback
from typing import Any, Dict, Type, TypeVar  # if this fails under Python 3.5.1, upgrade to 3.5.2  # noqa
# ... under Python 3.4/earlier, "typing" is an external module installed with
# pip, but under Python 3.5 it's build in to the standard library. However,
# early versions of Python 3.5's built-in "typing" don't include Type; see
# https://github.com/python/mypy/issues/1838

import arrow
from PyQt5.QtCore import (
    QObject,
    Qt,
    QThread,
    QTimer,
    pyqtSignal,
    pyqtSlot,
)
import serial
from serial import Serial
# pySerial 3: http://pyserial.readthedocs.org/en/latest/pyserial.html
from whisker.qt import exit_on_exception, StatusMixin

from starfeeder.constants import ThreadOwnerState

log = logging.getLogger(__name__)


CR = b'\r'
CRLF = b'\r\n'
LF = b'\n'
LF_STR = '\n'
NO_BYTES = b''

DEBUG_WRITE_TIMING = False

READ_TIMEOUT_SEC = 0.01
WRITE_TIMEOUT_SEC = 5.0  # None for blocking writes
# ... but blocking writes can cause the app to freeze if the serial port breaks
INTER_BYTE_TIMEOUT_SEC = None


# class SerialReceiveMessage(object):
#     def __init__(self, data: bytes, timestamp: arrow.Arrow) -> None:
#         self.data = data
#         self.timestamp = timestamp


class SerialReader(QObject, StatusMixin):  # separate reader thread
    """
    Object to monitor input from a serial port.
    Assigned to a thread (see below).
    """
    started = pyqtSignal()
    finished = pyqtSignal()
    # line_received = pyqtSignal(bytes, arrow.Arrow)  # problems
    # line_received = pyqtSignal(SerialReceiveMessage)  # fine?
    line_received = pyqtSignal(str, arrow.Arrow)

    def __init__(self, name: str = '?', parent: QObject = None,
                 eol: bytes = LF, input_encoding: str = 'ascii',
                 **kwargs) -> None:
        super().__init__(parent=parent, name=name, logger=log, **kwargs)
        self.serial_port = None  # set later
        self.eol = eol
        self.input_encoding = input_encoding

        self.len_eol = len(eol)
        self.finish_requested = False
        self.residual = b''

    @pyqtSlot(Serial)
    @exit_on_exception
    def start(self, serial_port: Serial) -> None:
        self.serial_port = serial_port
        self.debug("SerialReader starting. Using port {}".format(serial_port))
        self.finish_requested = False
        try:
            while not self.finish_requested:
                # We could use self.serial_port.readline(), noting also
                # http://pyserial.readthedocs.org/en/latest/shortintro.html
                # But we have to do some EOL stripping anyway, so let's just
                # do it in the raw.
                data = self.serial_port.read(1)  # will wait until timeout
                # ... will return b'' if no data
                data += self.serial_port.read(self.serial_port.inWaiting())
                if len(data) > 0:
                    self.process_data(data)
        # except serial.SerialException as e:
        except Exception as e:  # serial port errors
            self.error("Serial port error: {}; stopping".format(str(e)))
            self.error(traceback.format_exc())
        # - We should catch serial.SerialException, but there is a bug in
        #   serial/serialposix.py that does "if e[0] != errno.EAGAIN", which
        #   raises "TypeError: 'SerialException' object does not support
        #   indexing", so we use "except Exception as e:"
        # - 2016-11-28: looks like this bug is now fixed in PySerial 3.2.1
        # - ... but we can also get "termios.error: (22, 'Invalid argument'),
        #   e.g. ?if we use parity E with a socat port. So catch everything.
        self.finish()

    def process_data(self, data: bytes) -> None:
        """
        Adds the incoming data to any stored residual, splits it into lines,
        and sends each line on to the receiver.
        """
        self.debug("data: {}".format(repr(data)))
        timestamp = arrow.now()
        data = self.residual + data
        fragments = data.split(self.eol)
        lines = fragments[:-1]
        self.residual = fragments[-1]
        for line in lines:
            self.debug("line: {}".format(repr(line)))
            # self.line_received.emit(line, timestamp)
            # self.line_received.emit(b'Z5A2080A70C2C0001', timestamp)
            # msg = SerialReceiveMessage(data=line, timestamp=timestamp)
            # self.line_received.emit(msg)
            try:
                decoded = line.decode(self.input_encoding)
            except UnicodeDecodeError:
                self.critical("Received input that {} won't decode, ignoring: "
                              "{}".format(self.input_encoding, repr(data)))
                continue
            self.line_received.emit(decoded, timestamp)

    @pyqtSlot()
    @exit_on_exception
    def stop(self) -> None:
        """
        Request that the serial handler terminates.

        - If you call this using the default Signal/Slot mechanism, the
          connection type is QtCore.Qt.AutoConnection. When used across
          threads, this acts as Qt.QueuedConnection, sending from one thread
          and receiving on another. Atomic access to the variable is therefore
          not an issue, but it has to wait for the receiver to return to its
          message loop. So it won't work for the kind of loop we're operating
          here.
        - We could call the function directly from the calling thread, without
          any signal/slot stuff. That means we have to be sure the variable
          access is atomic.
        - We could use a signal/slot, but specify a Qt.DirectConnection.
          Then the function executes in the caller's thread, so we have to
          think about atomic access again.

        - What about atomic access?
          Well, there's QtCore.QMutex.

        - However, in this specific situation, of a Boolean variable that
          we are writing once, only reading from the other thread, and
          the exact timing of the off-to-on transition is immaterial, we
          should be fine here to ignore it.

        - In general, note also that the thread control mechanisms are LESS
          POWERFUL than you might guess. For example, you will never get a
          "started" message back from a thread function that loops
          indefinitely. (Therefore, if you want to start things in sequence,
          start writers before readers.)

        http://doc.qt.io/qt-4.8/qt.html#ConnectionType-enum
        https://srinikom.github.io/pyside-docs/PySide/QtCore/QMutex.html
        """
        # This is called via the Signal/Slot mechanism, which (when using the
        # default Auto Connection), executes the slot in the receiver's thread.
        # In other words, we don't have to use a mutex on this variable because
        # we are accessing it via the Qt signal/slot mechanism.
        self.debug("stopping")
        self.finish_requested = True

    def finish(self) -> None:
        self.finished.emit()


class SerialWriter(QObject, StatusMixin):  # separate writer thread
    """
    Object to send to a serial port.
    Assigned to a thread (see below).
    """
    started = pyqtSignal()
    finished = pyqtSignal()

    def __init__(self, name: str = '?',
                 output_encoding: str = 'utf8', eol: bytes = LF,
                 parent: QObject = None,
                 **kwargs) -> None:
        # ... UTF8 is ASCII for normal characters.
        super().__init__(parent=parent, name=name, logger=log, **kwargs)
        self.serial_port = None  # set later
        self.eol = eol
        self.output_encoding = output_encoding

        self.callback_timer = QTimer(self)
        # If you use QTimer() rather than QTimer(self), you get
        # "Timers cannot be started from another thread"
        # (under Windows; Linux Qt seems not to care), unless you manually
        # move the timer with a moveToThread() override (which is silly).
        # noinspection PyUnresolvedReferences
        self.callback_timer.timeout.connect(self.process_queue)
        self.queue = deque()  # contains (data, delay_ms) tuples
        # ... left end = next one to be sent; right = most recent request
        self.busy = False
        # = "don't send new things immediately; we're in a delay"

    @pyqtSlot(Serial)
    @exit_on_exception
    def start(self, serial_port: Serial) -> None:
        self.debug("starting")
        self.serial_port = serial_port
        self.started.emit()

    @pyqtSlot(str, int)
    @exit_on_exception
    def send(self, data: str, delay_ms: int) -> None:
        """
        Sending interface offered to others.
        We maintain an orderly output queue and allow delays.
        """
        try:
            encoded_bytes = data.encode(self.output_encoding)
        except UnicodeDecodeError:
            self.critical("Trying to send str that {} won't encode, ignoring: "
                          "{}".format(self.output_encoding, repr(data)))
            return
        self.queue.append((encoded_bytes, delay_ms))
        if not self.busy:
            self.process_queue()

    @pyqtSlot()
    @exit_on_exception
    def process_queue(self) -> None:
        """Deals with what's in the queue."""
        self.busy = False
        while len(self.queue) > 0:
            data, delay_ms = self.queue.popleft()
            if delay_ms > 0:
                # Put it back at the front of the queue with zero delay,
                # and wait for the delay before continuing.
                self.queue.appendleft((data, 0))
                self.wait(delay_ms)
                return
            self._send(data)

    def wait(self, delay_ms: int) -> None:
        self.busy = True
        self.callback_timer.setSingleShot(True)
        self.callback_timer.start(delay_ms)

    def _send(self, data: bytes) -> None:
        """
        Internal function to send data to the serial port directly.
        """
        try:
            outdata = data + self.eol
            self.debug("sending: {}".format(repr(outdata)))
            if DEBUG_WRITE_TIMING:
                t1 = arrow.utcnow()
            self.serial_port.write(outdata)
            # ... will raise SerialTimeoutException if a write timeout is
            #     set and exceeded
            self.serial_port.flush()
            if DEBUG_WRITE_TIMING:
                t2 = arrow.utcnow()
                nbytes = len(outdata)
                # noinspection PyUnboundLocalVariable
                microsec = (t2 - t1).microseconds
                self.debug(
                    "Sent {} bytes in {} microseconds ({} microseconds per "
                    "byte)".format(nbytes, microsec, microsec / nbytes))
        except Exception as e:  # serial port errors
            self.error("Serial port error: {}; continuing".format(str(e)))
            self.error(traceback.format_exc())


class SerialController(QObject, StatusMixin):  # separate controller thread
    """
    Does the thinking. Has its own thread.
    """
    data_send_requested = pyqtSignal(str, int)
    finished = pyqtSignal()

    def __init__(self, name: str, output_encoding: str = 'utf8',
                 parent: QObject = None, **kwargs) -> None:
        super().__init__(parent=parent, name=name, logger=log, **kwargs)
        self.output_encoding = output_encoding

    @pyqtSlot(str, arrow.Arrow)
    def on_receive(self, data: str, timestamp: arrow.Arrow) -> None:
        """Should be overridden."""
        pass

    @pyqtSlot()
    def on_start(self) -> None:
        """Should be overridden."""
        pass

    def send_str(self, data: str, delay_ms: int = 0) -> None:
        self.data_send_requested.emit(data, delay_ms)

    def send_bytes(self, data: bytes, delay_ms: int = 0) -> None:
        # This is a bit silly, but we're trying to avoid sending bytes objects
        # via PyQt5 signals; they seem to be vulnerable to corruption; see
        # README.rst
        try:
            data_str = data.decode(self.output_encoding)
        except UnicodeDecodeError:
            self.critical("Trying to send bytes that {} won't decode, ignoring"
                          ": {}".format(self.output_encoding, repr(data)))
            return
        self.data_send_requested.emit(data_str, delay_ms)

    def stop(self) -> None:
        """
        Can be overridden, for shutdown tasks.
        But you must emit the finished event when you're happy to proceed.
        """
        self.finished.emit()

    def report_status(self) -> None:
        """Should be overridden."""
        pass


SR = TypeVar('SR', bound='SerialReader')
SW = TypeVar('SW', bound='SerialWriter')
SC = TypeVar('SC', bound='SerialController')


class SerialOwner(QObject, StatusMixin):  # GUI thread
    """
    Encapsulates a serial port + reader (with thread) + writer (with thread)
    and the associated signals/slots.
    """
    # Outwards, to world:
    started = pyqtSignal()
    finished = pyqtSignal()
    state_change = pyqtSignal(int, str)
    # Inwards, to possessions:
    reader_start_requested = pyqtSignal(serial.Serial)
    writer_start_requested = pyqtSignal(serial.Serial)
    reader_stop_requested = pyqtSignal()
    writer_stop_requested = pyqtSignal()
    controller_stop_requested = pyqtSignal()
    status_requested = pyqtSignal()

    # noinspection PyUnresolvedReferences
    def __init__(self,
                 serial_args: Dict[str, Any],
                 parent: QObject = None,
                 rx_eol: bytes = LF,
                 tx_eol: bytes = LF,
                 callback_id: int = None,
                 name: str = '?',
                 input_encoding: str = 'ascii',  # serial device to us
                 output_encoding: str = 'utf8',  # us to serial device
                 reader_class: Type[SR] = SerialReader,
                 reader_kwargs: Dict[str, Any] = None,
                 writer_class: Type[SW] = SerialWriter,
                 writer_kwargs: Dict[str, Any] = None,
                 controller_class: Type[SC] = SerialController,
                 controller_kwargs: Dict[str, Any] = None,
                 read_timeout_sec: float = READ_TIMEOUT_SEC,
                 write_timeout_sec: float = WRITE_TIMEOUT_SEC,
                 inter_byte_timeout_sec: float = INTER_BYTE_TIMEOUT_SEC,
                 **kwargs):
        """
        serial_args: as per PySerial:
            port
            baudrate
            bytesize
            parity
            stopbits
            xonxoff
            rtscts
            dsrdtr
        """
        super().__init__(parent=parent, name=name, logger=log, **kwargs)
        self.callback_id = callback_id
        reader_kwargs = reader_kwargs or {}  # type: Dict[str, Any]
        writer_kwargs = writer_kwargs or {}  # type: Dict[str, Any]
        controller_kwargs = controller_kwargs or {}  # type: Dict[str, Any]
        self.serial_port = None  # type: Serial
        self.readerthread = QThread(self)
        self.writerthread = QThread(self)
        self.controllerthread = QThread(self)
        self.state = ThreadOwnerState.stopped

        # Serial port
        self.serial_args = serial_args
        self.serial_args.update(dict(
            timeout=read_timeout_sec,
            write_timeout=write_timeout_sec,
            inter_byte_timeout=inter_byte_timeout_sec,
        ))
        # timeout:
        #   read timeout... in seconds, when numeric
        #   See http://pyserial.readthedocs.org/en/latest/pyserial_api.html
        #   Low values: thread more responsive to termination; more CPU.
        #   High values: the converse.
        #   None: wait forever.
        #   0: non-blocking (avoid here).
        # inter_byte_timeout (formerly inter_byte_timeout):
        #   governs when read() calls return when there is a sufficient time
        #   between incoming bytes;
        #   https://github.com/pyserial/pyserial/blob/master/serial/serialposix.py  # noqa
        #   http://www.unixwiz.net/techtips/termios-vmin-vtime.html
        self.debug("Creating SerialOwner: {}".format(serial_args))

        # Serial reader/writer/controller objects
        reader_kwargs.setdefault('name', name)
        reader_kwargs.setdefault('eol', rx_eol)
        reader_kwargs.setdefault('input_encoding', input_encoding)
        self.reader = reader_class(**reader_kwargs)
        writer_kwargs.setdefault('name', name)
        writer_kwargs.setdefault('eol', tx_eol)
        self.writer = writer_class(**writer_kwargs)
        controller_kwargs.setdefault('name', name)
        controller_kwargs.setdefault('output_encoding', output_encoding)
        self.controller = controller_class(**controller_kwargs)

        # Assign objects to thread
        self.reader.moveToThread(self.readerthread)
        self.writer.moveToThread(self.writerthread)
        self.controller.moveToThread(self.controllerthread)

        # Connect object and thread start/stop events
        # ... start sequence
        self.writerthread.started.connect(self.writerthread_started)
        self.writer_start_requested.connect(self.writer.start)
        self.writer.started.connect(self.writer_started)
        self.readerthread.started.connect(self.readerthread_started)
        self.reader_start_requested.connect(self.reader.start)
        self.controllerthread.started.connect(self.controllerthread_started)
        self.started.connect(self.controller.on_start)

        # ... stop
        self.controller_stop_requested.connect(self.controller.stop)
        self.controller.finished.connect(self.controllerthread.quit)
        self.controllerthread.finished.connect(self.controllerthread_finished)
        self.controllerthread.finished.connect(self.reader_stop_requested)
        self.controllerthread.finished.connect(self.writer_stop_requested)
        self.reader_stop_requested.connect(self.reader.stop,
                                           Qt.DirectConnection)  # NB!
        self.reader.finished.connect(self.readerthread.quit)
        self.readerthread.finished.connect(self.readerthread_finished)
        self.writer_stop_requested.connect(self.writerthread.quit)
        self.writerthread.finished.connect(self.writerthread_finished)

        # Connect the status events
        self.reader.status_sent.connect(self.status_sent)
        self.reader.error_sent.connect(self.error_sent)
        self.writer.status_sent.connect(self.status_sent)
        self.writer.error_sent.connect(self.error_sent)
        self.controller.status_sent.connect(self.status_sent)
        self.controller.error_sent.connect(self.error_sent)
        self.status_requested.connect(self.controller.report_status)

        # Connect the control events
        self.reader.line_received.connect(self.controller.on_receive)
        self.controller.data_send_requested.connect(self.writer.send)

    # -------------------------------------------------------------------------
    # General state control
    # -------------------------------------------------------------------------

    def is_running(self) -> bool:
        running = self.state != ThreadOwnerState.stopped
        self.debug("is_running: {} (state: {})".format(running,
                                                       self.state.name))
        return running

    def set_state(self, state: ThreadOwnerState) -> None:
        self.debug("state: {} -> {}".format(self.state.name, state.name))
        self.state = state
        self.state_change.emit(self.callback_id, state.name)

    # -------------------------------------------------------------------------
    # Starting
    # -------------------------------------------------------------------------
    # We must have control over the order. We mustn't call on_start()
    # until BOTH threads have started. Therefore we must start them
    # sequentially, not simultaneously. The reader does not exit, so can't
    # notify us that it's started (since such notifications are via the Qt
    # message loop); therefore, we must start the writer first.

    def start(self) -> None:
        self.status("Starting serial")
        if self.state != ThreadOwnerState.stopped:
            self.error("Can't start: state is: {}".format(self.state.name))
            return
        self.set_state(ThreadOwnerState.starting)
        self.info("Opening serial port: {}".format(self.serial_args))
        try:
            self.serial_port = serial.Serial(**self.serial_args)
        except Exception as e:  # serial port errors
            self.error("Serial port error: {}; stopping".format(str(e)))
            self.error(traceback.format_exc())
            self.stop()
            return
        self.debug("starting writer thread")
        self.writerthread.start()

    @pyqtSlot()
    @exit_on_exception
    def writerthread_started(self) -> None:
        self.writer_start_requested.emit(self.serial_port)

    @pyqtSlot()
    @exit_on_exception
    def writer_started(self) -> None:
        self.debug("start: starting reader thread")
        self.readerthread.start()

    @pyqtSlot()
    @exit_on_exception
    def readerthread_started(self) -> None:
        self.reader_start_requested.emit(self.serial_port)
        # We'll never get a callback from that; it's now busy.
        self.debug("start: starting controller thread")
        self.controllerthread.start()

    @pyqtSlot()
    @exit_on_exception
    def controllerthread_started(self) -> None:
        self.set_state(ThreadOwnerState.running)
        self.started.emit()

    # -------------------------------------------------------------------------
    # Stopping
    # -------------------------------------------------------------------------
    # The stop sequence is more fluid, to cope with any problems.

    def stop(self) -> None:
        if self.state == ThreadOwnerState.stopped:
            self.error("Can't stop: state is: {}".format(self.state.name))
            return
        self.set_state(ThreadOwnerState.stopping)
        self.debug("stop: asking threads to finish")
        if self.check_everything_finished():
            return
        self.controller_stop_requested.emit()

    @pyqtSlot()
    @exit_on_exception
    def reader_finished(self) -> None:
        self.debug("SerialController.reader_finished")
        self.readerthread.quit()

    @pyqtSlot()
    @exit_on_exception
    def readerthread_finished(self) -> None:
        self.debug("stop: reader thread stopped")
        self.check_everything_finished()

    @pyqtSlot()
    @exit_on_exception
    def writerthread_finished(self) -> None:
        self.debug("stop: writer thread stopped")
        self.check_everything_finished()

    @pyqtSlot()
    @exit_on_exception
    def controllerthread_finished(self) -> None:
        self.debug("stop: controller thread stopped")
        self.check_everything_finished()

    def check_everything_finished(self) -> bool:
        if (self.readerthread.isRunning() or
                self.writerthread.isRunning() or
                self.controllerthread.isRunning()):
            return False
        # If we get here: yes, everything is finished. Tidy up.
        if self.serial_port:
            self.debug("Closing serial port")
            self.serial_port.close()  # for Windows
        self.set_state(ThreadOwnerState.stopped)
        self.finished.emit()
        return True

    # -------------------------------------------------------------------------
    # Info
    # -------------------------------------------------------------------------

    @pyqtSlot()
    def report_status(self) -> None:
        self.status("state: {}".format(self.state.name))
        # I think it's OK to request serial port information from the GUI
        # thread...
        try:
            sp = self.serial_port
            paritybits = 0 if sp.parity == serial.PARITY_NONE else 1
            n_bits = sp.bytesize + paritybits + sp.stopbits
            time_per_char_microsec = n_bits * 1000000 / sp.baudrate

            portset = OrderedDict()
            portset['name'] = sp.name
            portset['port'] = sp.port
            portset['baudrate'] = sp.baudrate
            portset['bytesize'] = sp.bytesize
            portset['parity'] = sp.parity
            portset['stopbits'] = sp.stopbits
            portset['timeout'] = sp.timeout
            portset['xonxoff'] = sp.xonxoff
            portset['rtscts'] = sp.rtscts
            portset['dsrdtr'] = sp.dsrdtr
            portset['write_timeout'] = sp.write_timeout
            portset['inter_byte_timeout'] = sp.inter_byte_timeout
            self.status("Serial port settings: " + ", ".join(
                "{}={}".format(k, v) for k, v in portset.items()))

            portinfo = OrderedDict()
            portinfo['in_waiting'] = sp.in_waiting
            if platform.system() in ['Linux', 'Windows']:
                portinfo['out_waiting'] = sp.out_waiting
            portinfo['break_condition'] = sp.break_condition
            portinfo['rts'] = sp.rts
            portinfo['cts'] = sp.cts
            portinfo['dtr'] = sp.dtr
            portinfo['dsr'] = sp.dsr
            portinfo['ri'] = sp.ri
            portinfo['cd'] = sp.cd
            portinfo['rs485_mode'] = sp.rs485_mode
            portinfo['[time_per_char_microsec]'] = time_per_char_microsec
            self.status("Serial port info: " + ", ".join(
                "{}={}".format(k, v) for k, v in portinfo.items()))
        except Exception as e:  # serial port errors
            self.warning(
                "Serial port is unhappy - may be closed, or trying to read a "
                "non-existent attribute; error: {}".format(str(e)))
            self.warning(traceback.format_exc())

        self.status_requested.emit()



#!/usr/bin/env python
# starfeeder/task.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging

import arrow
# from whisker.debug_qt import debug_object
from whisker.sqlalchemy import session_thread_scope
from whisker.qt import exit_on_exception
from whisker.qtclient import WhiskerTask

from PyQt5.QtCore import QObject, pyqtSlot
from starfeeder.models import (
    MassEvent,  # for type hints
    MassEventRecord,
    MasterConfig,
    RfidEvent,  # for type hints
    RfidEventRecord,
)
from starfeeder.settings import get_database_settings
from starfeeder.version import VERSION

log = logging.getLogger(__name__)


class WeightWhiskerTask(WhiskerTask):  # Whisker thread B
    """Doesn't define an end, deliberately."""

    def __init__(self, wcm_prefix: str = "", parent: QObject = None,
                 name: str = "whisker_task", **kwargs) -> None:
        super().__init__(parent=parent, name=name, **kwargs)
        self.wcm_prefix = wcm_prefix
        self.dbsettings = get_database_settings()
        self.rfid_effective_time_s = None

    @pyqtSlot()
    def thread_started(self) -> None:
        with session_thread_scope(self.dbsettings) as session:
            config = MasterConfig.get_singleton(session)
            self.rfid_effective_time_s = config.rfid_effective_time_s

    @pyqtSlot()
    @exit_on_exception
    def on_connect(self) -> None:
        # self.debug("DERIVED on_connect")
        # debug_object(self)
        # self.whisker.command("TimerSetEvent 2000 5 bop")
        self.whisker.report_name("Starfeeder {}".format(VERSION))

    @pyqtSlot(str, arrow.Arrow, int)
    @exit_on_exception
    def on_event(self, event, timestamp, whisker_timestamp_ms):
        pass
        # if event == "bop":
        #     self.status("boop")

    def broadcast(self, msg: str) -> None:
        if self.wcm_prefix:
            msg = "{}{}".format(self.wcm_prefix, msg)
        self.whisker.broadcast(msg)

    @pyqtSlot(RfidEvent)
    @exit_on_exception
    def on_rfid(self, rfid_event: RfidEvent) -> None:
        """
        Record an RFID event.

        Since this task runs in a non-GUI thread, it's a good place to do the
        main RFID processing.

        Only one thread should be writing to the database, to avoid locks.

        Don't hold the session too long, on general principles.
        """
        if not isinstance(rfid_event, RfidEvent):
            log.critical("Bad rfid_event: {}".format(rfid_event))
            return
        with session_thread_scope(self.dbsettings) as session:
            RfidEventRecord.record_rfid_detection(
                session, rfid_event, self.rfid_effective_time_s)
        # self.status("RFID received: {}".format(rfid_event))
        if self.whisker.is_connected():
            self.broadcast(
                "RFID_EVENT: reader {reader}, RFID {rfid}, "
                "timestamp {timestamp}".format(
                    rfid=rfid_event.rfid,
                    reader=rfid_event.reader_name,
                    timestamp=rfid_event.timestamp,
                )
            )

    @pyqtSlot(MassEvent)
    @exit_on_exception
    def on_mass(self, mass_event: MassEvent) -> None:
        """
        Receive a mass event. Ask the MassIdentifiedEvent class to work out if
        it represents an identified mass event (and store it, if so).
        Broadcast the information to the Whisker client.
        """
        if not isinstance(mass_event, MassEvent):
            log.critical("Bad mass_event: {}".format(mass_event))
            return
        if not mass_event.locked or mass_event.rfid is None:
            return
        with session_thread_scope(self.dbsettings) as session:
            MassEventRecord.record_mass_detection(session, mass_event)
        if self.whisker.is_connected():
            self.broadcast(
                "MASS_EVENT: reader {reader}, RFID {rfid}, balance {balance}, "
                "mass {mass_kg} kg, timestamp {timestamp}".format(
                    reader=mass_event.reader_name,
                    rfid=mass_event.rfid,
                    balance=mass_event.balance_name,
                    mass_kg=mass_event.mass_kg,
                    timestamp=mass_event.timestamp,
                )
            )
